@@ -6,8 +6,15 @@
 #ifndef _BIRD_CONFIG_H_
 #define _BIRD_CONFIG_H_
 
+#define XSTR2(X) #X
+#define XSTR1(X) XSTR2(X)
+
 /* BIRD version */
-#define BIRD_VERSION "1.6.8"
+#ifdef GIT_LABEL
+#define BIRD_VERSION XSTR1(GIT_LABEL)
+#else
+#define BIRD_VERSION "2.0.8"
+#endif
 
 /* Include parameters determined by configure script */
 #include "sysdep/autoconf.h"
